extends Node
class_name SSTVEncoder

const TWO_PI = PI * 2
const SAMPLE_RATE = 48000.0
const VIS_BIT_FREQ = {
	'ONE': 1100.0,
	'ZERO': 1300.0,
}
const PREFIX_PULSE_LENGTH = 0.1  # 100 ms
const HEADER_PULSE_LENGTH = 0.3  # 300 ms
const HEADER_BREAK_LENGTH = 0.01 # 10 ms
const VIS_BIT_LENGTH = 0.03      # 30 ms
const SYNC_PULSE_FREQ = 1200.0
const BLANKING_PULSE_FREQ = 1500.0

var outputFloats = PackedFloat32Array()

func EncodePrefix():
	outputFloats.append_array(GenerateTone(1900, PREFIX_PULSE_LENGTH))
	outputFloats.append_array(GenerateTone(1500, PREFIX_PULSE_LENGTH))
	outputFloats.append_array(GenerateTone(1900, PREFIX_PULSE_LENGTH))
	outputFloats.append_array(GenerateTone(1500, PREFIX_PULSE_LENGTH))
	outputFloats.append_array(GenerateTone(2300, PREFIX_PULSE_LENGTH))
	outputFloats.append_array(GenerateTone(1500, PREFIX_PULSE_LENGTH))
	outputFloats.append_array(GenerateTone(2300, PREFIX_PULSE_LENGTH))
	outputFloats.append_array(GenerateTone(1500, PREFIX_PULSE_LENGTH))

func EncodeHeader(vis: Array):
	outputFloats.append_array(GenerateTone(1900, HEADER_PULSE_LENGTH))
	outputFloats.append_array(GenerateTone(SYNC_PULSE_FREQ, HEADER_BREAK_LENGTH))
	outputFloats.append_array(GenerateTone(1900, HEADER_PULSE_LENGTH))

	# VIS Code
	outputFloats.append_array(GenerateTone(SYNC_PULSE_FREQ, VIS_BIT_LENGTH))

	var parity = 0
	vis.reverse()
	for bit in vis:
		var bitFreq = VIS_BIT_FREQ["ONE"] if bit else VIS_BIT_FREQ["ZERO"]
		if bit:
			parity += 1
		outputFloats.append_array(GenerateTone(bitFreq, VIS_BIT_LENGTH))

	var parityFreq = VIS_BIT_FREQ["ZERO"] if parity % 2 == 0 else VIS_BIT_FREQ["ONE"]
	outputFloats.append_array(GenerateTone(parityFreq, VIS_BIT_LENGTH))

	outputFloats.append_array(GenerateTone(SYNC_PULSE_FREQ, VIS_BIT_LENGTH))

var phase = 0.0

func GenerateTone(frequency: float, duration: float) -> PackedFloat32Array:
	var samples = PackedFloat32Array()
	var totalSamples = int(duration * SAMPLE_RATE)
	
	var timeStep = 1.0 / SAMPLE_RATE
	for i in range(totalSamples):
		phase += TWO_PI * frequency * timeStep
		var sample = sin(phase)
		samples.append(sample)
	
	return samples

func GenerateToneFromCurve(frequencies: Array, duration: float) -> PackedFloat32Array:
	var samples = PackedFloat32Array()
	var totalSamples = int(duration * SAMPLE_RATE)
	var points = frequencies.size()

	var timeStep = 1.0 / SAMPLE_RATE

	for i in range(totalSamples):
		var time = float(i) * timeStep
		var normalizedTime = time / duration
		var indexx = normalizedTime * (points - 1)
		var index = int(indexx)
		var frac = indexx - index
		var freq: float
		
		if index >= points - 1:
			freq = frequencies[points - 1]
		else:
			freq = lerp(frequencies[index], frequencies[index + 1], frac)
		
		phase += TWO_PI * freq * timeStep

		var sample = sin(phase)
		samples.append(sample)

	return samples

func EncodeSSTV(image: Image, skipHeader: bool = false):
	pass

func SaveAsWav(path: String):
	var file = FileAccess.open(path, FileAccess.WRITE)
	var outputData = PackedByteArray()

	var samples = outputFloats.size()
	var chunkSize = samples * 2
	var fileSize = 44 + chunkSize - 8

	# WAV Header
	outputData.append_array("RIFF".to_ascii_buffer())
	outputData.append_array(ConvertEndian(fileSize, 4))
	outputData.append_array("WAVEfmt ".to_ascii_buffer())
	outputData.append_array(ConvertEndian(16, 4)) # PCM header size
	outputData.append_array(ConvertEndian(1, 2))  # PCM format
	outputData.append_array(ConvertEndian(1, 2))
	outputData.append_array(ConvertEndian(SAMPLE_RATE, 4))
	outputData.append_array(ConvertEndian(SAMPLE_RATE * 1 * 16 / 8, 4)) # Byte rate
	outputData.append_array(ConvertEndian(1 * 16 / 8, 2)) # Block align
	outputData.append_array(ConvertEndian(16, 2))
	outputData.append_array("data".to_ascii_buffer())
	outputData.append_array(ConvertEndian(chunkSize, 4))

	# Sample data
	for sample in outputFloats:
		var signed16Sample = int(clamp(sample * 32767, -32767, 32767))
		outputData.append_array(ConvertEndian(signed16Sample, 2))

	file.store_buffer(outputData)
	file.close()

func ConvertEndian(value: int, bytes: int) -> PackedByteArray:
	var reversedBytes = PackedByteArray()
	for i in range(bytes):
		reversedBytes.append((value >> (8 * i)) & 0xFF)
	return reversedBytes
